/***********************************************************/
/* Author: Kyri Pavlou         16 January 2009             */
/* IMPORTANT: This implementation of the Tiled-Bitmap      */
/* algorithm is an improved variant of the one described   */
/* in the TKDE submission "The Tiled Bitmap Forensic       */
/* Analysis Algorithm." It employs a series of cumulative  */
/* (termed black) hash chains which are computed           */
/* in real-time (instead of at each validation scan).      */
/* These black hash chains are common to all the forensic  */   
/* analysis algorithms (monochromatic, RGBY, Tiled Bitmap, */
/* a3D) and allows for the fast location of the interface  */
/* between the tamper-free area and potentially corrupted  */ 
/* areas using binary search.                              */
/***********************************************************/
#include "algorithms.h"

/* Global variables */
static int c_logical = ZERO;  		/*the logical size of C_set*/
static int c_capacity = INIT_CAP;      	/*the physical size of C_set*/

/***************************************************************/
/*Input: 						       */
/*	  t_FVF: time of first validation failure	       */
/*  	    I_N: the notarization interval 		       */
/*	      k: the dimension of the boolean domain  	       */
/*	         of the AND function  			       */
/* default_flag: set to TRUE if the value of k is not known    */
/*		 and a default value of 2 is used. It is FALSE */
/*               if the caller has definite knowledge of k     */
/*Output: the candidate set C_set with the corrupted granules  */
/*							       */
/*The  tiled-bitmap algorithm first locates t_RVS and then     */
/*linealry iterates through the tiles after t_RVS. At each     */
/*tile it checks validation result for the hash chain          */ 
/*(c_0) covering the entire tile. If it is TRUE is proceeds    */
/*to the next one. If not, it computes the target bit pattern  */
/*by concatenating the validation results of all the rest of   */
/*the constituent hash chains. This target bit pattern is      */
/*used by the functions in CandidateSetFastest.c to generate   */
/*the candidate set: the set C_Set of all possibly corrupted   */
/*granules.						       */
/***************************************************************/
void Tiled_Bitmap(int t_FVF, int I_N, int k, int default_flag)
{
    DBG(printf("Inside Tiled_Bitmap\n"));
    unsigned int target;   /*the numerical representation of the binary target number in decimal*/
    int t,i,j;
    int *C_set;
    int g;

    int *candidates;

    C_set = (int *)mymalloc(c_capacity*sizeof(int));
    int n = num_chains_per_tile - 1; /*number of c_1 -- c_x chains in a tile excluding c_0*/

    DBG(printf("The length of the target%d\n", n ));

    int t_RVS = find_t_RVS(t_FVF, I_N);
    int UTB = t_FVF;
    int LTB = MAX(t_FVF - I_V, t_RVS);

    t = t_RVS+1; /*Important: t_RVS does not return the 0th granule of the current tile*/
                 /*it returns the (N-1)th of the previous tile*/
    DBG(printf("t_RVS is %d\n", t_RVS));
    while(t < t_FVF) /*iterate through all tiles after t_RVS*/
    {
        if(val_check(t,0) == FALSE)
        {
            DBG(printf("c_0 evaluated to FALSE\n"));
            target = 0;

            for( i = n; i >= 1; i--)
            {   /*target bit pattern calculated from the tiles array*/
                target = target + (1 << (n-i))*val_check(t, i);
                DBG(printf(" the target is %d\n",target));
            } 
            DBG(printf("The target is %d\n",target));
            candidateSetCached(target, n , k, default_flag, &candidates);	/*call to the CandidateSetFastest.c 
										function to generate the candidate set 
										for the current tile*/
            i = 0;
            /*for each granule in candidate set (candidates) use its relative position (within tile)
	      to find its global position*/
            while(i < logical_size) 
            {
                g = t + R_s*candidates[i];		
                DBG(printf("g is %d and t is %d \n", g, t));
                append(&C_set, g); /*accumulate all the potentially corrupted 
				   granules from each tile into the final C_set*/
                c_logical++;
                i++;
            }
            DBG(printf("Tiled_Bitmap: the logical size of C_set: %d\n", c_logical));
            logical_size = 0;
  	    free(candidates);
        }
        t = t + I_N;
    }

    DBG(printf("Final logical size of C_set = %d \n", c_logical));
    printf("%d < t_c <= %d\n", LTB, UTB);
    printf("***The C_set is***\n"); /*Output the C_set*/
    printf("C_set = {");
    for(j = 0; j < c_logical; j++)
    {
        if(R_s == 1)
            printf(" %d ",C_set[j]);
        else
            printf(" [%d, %d] ", C_set[j], C_set[j] + R_s -1);		
    }
    printf("}\n");
    free(C_set);
}

/***************************************************************/
/* Appends corrupted granule g to candidate set (C_set) for    */
/* the tiled bitmap algorithm.				       */ 
/* If there is not enough space in the set the array is        */
/* resized to double its previous capacity.                    */
/***************************************************************/
void append(int **C_set, int g)
{
    DBG(printf("Appending %d to candidate set...\n", g));
    if(c_logical < c_capacity)
    {
        (*C_set)[c_logical] = g;
    }
    else
    {
        DBG(printf("Resizing the candidate set...\n"));
        int i;
        int *C_temp;
        c_capacity *= 2;
        C_temp = (int *)mymalloc(c_capacity*sizeof(int));
        for(i = 0; i < c_logical; i++)
        {
            C_temp[i] = (*C_set)[i];
        }
        C_temp[c_logical] = g;
        free(*C_set);
        *C_set = C_temp;
    }
}

/******************************************************************/
/* Returns the validation result (T/F) of the partial hash chain  */
/* in tile t with relative position i (within the tile).    	  */
/******************************************************************/
int val_check(int t, int i)
{
    counter_tiled_f++;
    return tiles[(-1+(int)ceil((float)t / I_N))*num_chains_per_tile + i];
}

